<?php

namespace app\controllers;

use app\models\InOut;
use app\models\InOutDetail;
use app\models\InOutSearch;
use app\models\Param;
use app\models\Planning;
use Yii;
use yii\helpers\HtmlPurifier;
use yii\helpers\Json;
use yii\web\Controller;
use yii\web\NotFoundHttpException;
use yii\filters\VerbFilter;

/**
 * Class InOutController
 * @author Ing. Maryeiny Losada Romero
 * Ingeneo S.A.S
 * Controller to manage the Inputs and Outputs of Products in SIESA
 * @package app\controllers
 */
class InOutController extends Controller
{
    
    /**
     * @inheritdoc
     */
    public function behaviors()
    {
        return [
            'verbs' => [
                'class' => VerbFilter::className(),
                'actions' => [
                    'delete' => ['POST'],
                ],
            ],
        ];
    }

    /**
     * List of Inputs
     * @return string
     * @throws \yii\base\InvalidRouteException
     * @throws \yii\console\Exception
     */
    public function actionIndexIn()
    {
        $searchModel = new InOutSearch();
        $queryParams = array_merge(array(), Yii::$app->request->getQueryParams());
        $queryParams['InOutSearch']['type'] = InOut::TYPE_IN;
        $dataProvider = $searchModel->search($queryParams);
        $title = 'Input';

        // Visualización de los Inventarios habilitados
        return $this->render('index', [
            'searchModel' => $searchModel,
            'dataProvider' => $dataProvider,
            'title' => $title
        ]);
    }

    /**
     * List of Outputs
     * @return string
     * @throws \yii\base\InvalidRouteException
     * @throws \yii\console\Exception
     */
    public function actionIndexOut()
    {
        $searchModel = new InOutSearch();
        $queryParams = array_merge(array(), Yii::$app->request->getQueryParams());
        $queryParams['InOutSearch']['type'] = InOut::TYPE_OUT;
        $dataProvider = $searchModel->search($queryParams);
        $title = 'Output';

        // Visualización de los Inventarios habilitados
        return $this->render('index', [
            'searchModel' => $searchModel,
            'dataProvider' => $dataProvider,
            'title' => $title
        ]);
    }

    /**
     * Delete of Inputs and Outputs
     * @return array
     * @throws NotFoundHttpException
     * @throws \Throwable
     * @throws \yii\db\StaleObjectException
     */
    public function actionDelete($id){
        try{
            $model = $this->findModel($id);
            if($model->type == InOut::TYPE_IN){
                $redirect = 'index-in';
            }else{
                $redirect = 'index-out';
            }
            $model->delete();
            InOutDetail::deleteAll('in_out_id = '.$id);
            return $this->redirect([$redirect]);
        }catch (\Exception $e){
            Yii::$app->session->setFlash('error', 'Error al Eliminar la Entrada o Salida.');
        }
    }

    /**
     * Displays a InOut model.
     * @param integer $id
     * @return mixed
     * @throws NotFoundHttpException if the model cannot be found
     */
    public function actionView($id)
    {
        $status =  json_decode(Param::findOne(['key' => 'status_requisitions_in' ])->value, true);
        return $this->render('view', [
            'model' => $this->findModel($id),
            'status' => $status
        ]);
    }

    /**
     * Event Search
     * @param $id
     * @return Planning|null
     * @throws NotFoundHttpException
     */
    protected function findModel($id)
    {
        if (($model = InOut::findOne($id)) !== null) {
            return $model;
        }

        throw new NotFoundHttpException(Yii::t('app', 'The requested page does not exist.'));
    }

    /**
     * Creates a new Entry model.
     * If creation is successful, the browser will be redirected to the 'view' page.
     * @return mixed
     */
    public function actionCreateIn()
    {
        $model = new InOut();

        if ($model->load(Yii::$app->request->post())) {
            $model->type = InOut::TYPE_IN;
            $model->type_created = InOut::TYPE_CREATED_ADMID;
            $model->status = InOut::STATUS_PENDIENTE;
            $this->_postSave($model);
            return $this->render('create', [
                'model' => $model,
            ]);
        } else {
            return $this->render('create', [
                'model' => $model,
            ]);
        }
    }

    /**
     * Updates an existing Entry model.
     * If update is successful, the browser will be redirected to the 'view' page.
     * @param integer $id
     * @return mixed
     * @throws NotFoundHttpException if the model cannot be found
     */
    public function actionUpdate($id)
    {
        $model = $this->findModel($id);

        if ($model->load(Yii::$app->request->post())) {
            $this->_postSave($model);
            return $this->render('update', [
                'model' => $model,
            ]);
        } else {
            if($model->status == InOut::STATUS_ALLOW){
                return $this->redirect(['index-in']);
            }else{
                return $this->render('update', [
                    'model' => $model,
                ]);
            }
        }
    }

    /**
     * Metodo para guardar las compras
     *
     * @author Alexander Arcila <alexander.arcila@ingeneo.com.co>
     * @copyright 2017 INGENEO S.A.S.
     * @license http://www.opensource.org/licenses/bsd-license.php  BSD License
     * @version SVN: $Id$
     * @link http://www.ingeneo.com.co
     */
    private function _postSave(&$model){
        //Captura de datos Post ------------------------------------------------
        $msg = '';
        if (isset($_POST['InOut'])) {
            //Validacion de status segun botones -------------------------------
            if (isset($_POST['status']) && $_POST['status'] === InOut::STATUS_PENDIENTE) {
                $model->status = InOut::STATUS_PENDIENTE;
            } elseif (isset($_POST['status']) && $_POST['status'] === InOut::STATUS_ALLOW) {
                $model->status = InOut::STATUS_ALLOW;
            } else {
                $model->status = InOut::STATUS_PARCIAL;
            }
            //------------------------------------------------------------------
            $razonSocial = explode(' - ',$model->business_name);
            $model->nit = isset($razonSocial[0]) ? $razonSocial[0] : null;
            $model->business_name = isset($razonSocial[1]) ? $razonSocial[1] : null;
            $model->validate();
            //------------------------------------------------------------------
            if ($model->validate()) {
                $quantitys = $_POST['cantidad'];
                if (count($quantitys) > 0) {
                    // Guardar los datos del pedido ----------------------------
                    $model->save();
                    //----------------------------------------------------------
                    $error = true;
                    //Variable para indicar al vendedor en caso de no haber ----
                    InOutDetail::deleteAll(['in_out_id'=>$model->id]);
                    foreach ($quantitys as $key => $value) {
                        if (!is_numeric($value)) {
                            $value = 0;
                        }

                        if (!empty($value) && $value > 0) {
                            $error = false;
                            $undExist =  '';
                            $tempReferencia = explode('-',$key);
                            $referencia = $tempReferencia[1];
                            $description = HtmlPurifier::process($_POST[$key . '_description']);
                            $idCategory = HtmlPurifier::process($_POST[$key . '_cat']);
                            //--------------------------------------------------
                            if(isset($_POST[$key . '_inv'])){
                                $undExist = HtmlPurifier::process($_POST[$key . '_inv']);
                            }
                            //Guardamos las referencias del pedido -------------
                            $entryDetail = new InOutDetail();
                            $entryDetail->in_out_id = $model->id;
                            $entryDetail->reference = $referencia;
                            $entryDetail->description = $description;
                            $entryDetail->quantity = $value;
                            $entryDetail->category = $idCategory;
                            if(!$entryDetail->save()){
                                $error = false;
                            }
                        }
                    }
                    //----------------------------------------------------------
                    if ($error) {
                        Yii::$app->session->setFlash('danger',
                            'No hay cantidades ingresadas');
                        $model->delete();
                    } else {
                        //Envio de correo a los pedidos aprobados --------------
                        if ($model->status === InOut::STATUS_ALLOW) {
                            if(!$model->sendEntryERP($model)){
                                $model->status = InOut::STATUS_PENDIENTE;
                                $model->save();
                                $msg = [
                                    'tipo'=>'success',
                                    'msg'=>Yii::t('app', 'Error al guardar la Entrada.')
                                        . ' - No. PEDIDO: '.$model->id
                                ];
                            }else{
                                $msg = [
                                    'tipo'=>'success',
                                    'msg'=>Yii::t('app', 'Entrada guardada con éxito.')
                                        . ' - No. PEDIDO: '.$model->id
                                ];
                            }
                            return $this->redirect(['view', 'id' => $model->id, 'msg'=>$msg]);
                        }
                        //Notificacion al solicitar aprobacion -----------------
                        else if ($model->status === InOut::STATUS_PENDIENTE) {
                            $msg = [
                                'tipo'=>'success',
                                'msg'=>Yii::t('app', 'Entrada guardada como '.InOut::STATUS_PENDIENTE)
                                    . ' - No. ENTRADA: '.$model->id
                            ];
                            return $this->redirect(['view', 'id' => $model->id, 'msg'=>$msg]);
                        }
                    }
                }
            }
        }
    }

    /**
     * Retorna las bodegas
     *
     * @param type $q
     * @param type $id
     *
     * @return json
     * @author Alexander Arcila <alexander.arcila@ingeneo.com.co>
     * @copyright 2018 INGENEO S.A.S.
     * @license http://www.opensource.org/licenses/bsd-license.php  BSD License
     * @version SVN: $Id$
     * @link http://www.ingeneo.com.co
     */
    public function actionGetinfopdv() {
        \Yii::$app->response->format = \yii\web\Response::FORMAT_JSON;
        if(isset($_POST['nit'])  && !empty($_POST['nit'])
            && isset($_POST['business_name'])  && !empty($_POST['business_name'])){
            $nit = HtmlPurifier::process($_POST['nit']);
            $tempBusinessName = HtmlPurifier::process($_POST['business_name']);
            $business_name = explode(' - ', $tempBusinessName);
            $requisition = new InOut();
            return $requisition->getInfopdv($nit, $business_name[1]);
        }else{
            return false;
        }
    }

    /**
     * Retnorna los botones
     *
     * @author Alexander Arcila <alexander.arcila@ingeneo.com.co>
     * @copyright 2018 INGENEO S.A.S.
     * @license http://www.opensource.org/licenses/bsd-license.php  BSD License
     * @version SVN: $Id$
     * @link http://www.ingeneo.com.co
     */
    public function actionChangebuttom() {
        if (isset($_POST['type']) && $_POST['type'] === 'liquidar') {
            return $this->renderAjax('_buttomLiquidar');
        } else {
            return $this->renderAjax('_buttomSend');
        }
    }
}
